#!/bin/bash

# Parallel Runner 
# ---------------

# This sbatch script can be used to run the same command/program against many
# input files, in parallel across multiple cluster nodes. It distributes the tasks
# evenly between cluster nodes.

# First, you need to write a separate script (TASK_SCRIPT) that will process a single
# input file, with the name of the input passed as an argument to that
# script. 
#
# e.g. running
#
#     process_single.sh myinput_abc123.txt
#
# would run all of the processing on the single input file 'myinput_abc123.txt'

# Second, customize the settings below, as described below so that we can run your
# processing in parallel. You will submit to the job scheduler with sbatch.
#
# e.g. running
#
#    sbatch process_parallel_sbatch.sh
#

# Contact biohpc-help@utsouthwestern.edu with any questions.

#----------------------------------------------------------------------------------------
# USER SETTINGS
#-----------------------------------------------------------------------------------------

# Name for this job
#SBATCH --job-name=parallelExample

# Use the super partition - any available node
#SBATCH --partition=super

# Number of nodes to use
#SBATCH --nodes=2

# Number of tasks to run at once across all nodes
#SBATCH --tasks=56

# How long this job can run
#SBATCH --time=7-00:00:00

# How many threads (cores) each task will use
CORES_PER_TASK=1

# Command that will generate a list of the inputs we want to process. This is usually
# 'ls' with some kind of wildcard to find files of interest. Check that the command lists
# the expected input files before you submit to the cluster.

# Here we assume we have image files within per-sample directories
INPUTS_COMMAND="ls BWimages"

# The Script we want to run on each input
TASK_SCRIPT="single.sh"


#----------------------------------------------------------------------------------------
# DO NOT EDIT BELOW THIS LINE
#-----------------------------------------------------------------------------------------

module load parallel

# We start each task under srun, so it will be distributed
# over our node allocation, and visible to slurm
# as a job step for easy monitoring and control.
# Each task runs on only 1 node (-N1), is started as a
# single process (-n1) and requires specified number of
# cores.
SRUN="srun --exclusive -n1 -N1 -c $CORES_PER_TASK"

# GNU parallel will co-ordinate running the tasks in
# parallel and avoid swamping the SLURM scheduler with all
# job steps at once. We tell it to run -j tasks
# at a time, have a small delay in-between tasks (for SLURM to catch up
# when a task finishes), and write the run-time and status etc into task.log.
PARALLEL="parallel --delay .2 -j $SLURM_NTASKS --joblog task.log"

# Let's show what we are going to do.
echo "Parallel Runner"
echo "---------------"
echo
echo "I will run the following script:"
echo "   $TASK_SCRIPT"
echo
echo "On the following input files (passed as 1st argument to the script)"
eval $INPUTS_COMMAND
echo
echo "I will use $SLURM_JOB_NUM_NODES nodes, to process these files."
echo "I will process up to $SLURM_NTASKS files at once, in parallel."
echo


# Do the work!
# Pipe our list of files into parallel
# Tell parallel to call srun for each file
# Tell srun to launch our task script for a single file each time it is called
# by parallel. The {} passes the input file name for the particular
# invocation.
eval $INPUTS_COMMAND | $PARALLEL $SRUN sh $TASK_SCRIPT {}

